#!/usr/bin/env python3

import click
from click import option, Option, UsageError
from pydamage.main import pydamage_analyze
from pydamage.citation import get_citation
from pydamage.filter import apply_filter
from pydamage import __version__
from collections import OrderedDict


class NaturalOrderGroup(click.Group):
    def __init__(self, name=None, commands=None, **attrs):
        super(NaturalOrderGroup, self).__init__(name=name, commands=None, **attrs)
        if commands is None:
            commands = OrderedDict()
        elif not isinstance(commands, OrderedDict):
            commands = OrderedDict(commands)
        self.commands = commands

    def list_commands(self, ctx):
        return self.commands.keys()


class MutuallyExclusiveOption(Option):
    # Credits goes to Stan Chang for this code snippet
    # https://gist.github.com/stanchan/bce1c2d030c76fe9223b5ff6ad0f03db

    def __init__(self, *args, **kwargs):
        self.mutually_exclusive = set(kwargs.pop("mutually_exclusive", []))
        help = kwargs.get("help", "")
        if self.mutually_exclusive:
            ex_str = ", ".join(self.mutually_exclusive)
            kwargs["help"] = help + (
                " NOTE: This argument is mutually exclusive with "
                " arguments: [" + ex_str + "]."
            )
        super(MutuallyExclusiveOption, self).__init__(*args, **kwargs)

    def handle_parse_result(self, ctx, opts, args):
        if self.mutually_exclusive.intersection(opts) and self.name in opts:
            raise UsageError(
                "Illegal usage: `{}` is mutually exclusive with "
                "arguments `{}`.".format(self.name, ", ".join(self.mutually_exclusive))
            )

        return super(MutuallyExclusiveOption, self).handle_parse_result(ctx, opts, args)


@click.group(cls=NaturalOrderGroup)
@click.version_option(__version__)
@click.pass_context
@click.option(
    "-o",
    "--outdir",
    type=click.Path(writable=True, dir_okay=True),
    default="pydamage_results",
    show_default=True,
    help="Output directory",
)
@click.option(
    "-t",
    "--threshold",
    default=0.5,
    type=float,
    show_default=True,
    help="Predicted accuracy filtering threshold. Set to 0 for finding threshold with kneed method",
)
def cli(ctx, outdir, threshold):
    """\b
    PyDamage: Damage parameter estimation for ancient DNA
    Author: Maxime Borry
    Contact: <borry[at]shh.mpg.de>
    Homepage & Documentation: github.com/maxibor/pydamage

    """

    ctx.ensure_object(dict)

    ctx.obj["outdir"] = outdir
    ctx.obj["threshold"] = threshold
    pass


@cli.command()
@click.pass_context
@click.argument("bam", type=click.Path(exists=True))
@click.option(
    "-w",
    "--wlen",
    default=13,
    type=int,
    show_default=True,
    help="Window length (in bp) for damage modeling",
)
@click.option(
    "-m",
    "--minlen",
    default=0,
    type=int,
    show_default=True,
    help="Minimum of length of reference sequence to consider",
)
@click.option(
    "-p",
    "--process",
    default=2,
    type=int,
    show_default=True,
    help="Number of processes for parallel computing",
)
@click.option(
    "-s", "--show_al", is_flag=True, help="Display alignments representations"
)
@click.option("-pl", "--plot", is_flag=True, help="Write damage fitting plots to disk")
@click.option("-vv", "--verbose", is_flag=True, help="Verbose mode")
@click.option(
    "-f", "--force", is_flag=True, help="Force overwriting of results directory"
)
@click.option(
    "-g",
    "--group",
    is_flag=True,
    help="Use entire BAM file as single reference for analysis "
    "(ignore reference headers)",
)
@click.option("-n", "--no_ga", is_flag=True, help="Do not use G->A transitions")
@click.option(
    "-r",
    "--rescale",
    is_flag=True,
    cls=MutuallyExclusiveOption,
    help="Rescale base quality scores using the PyDamage damage model",
    mutually_exclusive=["subsample"],
)
@click.option(
    "-s",
    "--subsample",
    cls=MutuallyExclusiveOption,
    type=click.FloatRange(0, 1),
    help="Subsample a fraction of the reads for damage modeling",
    mutually_exclusive=["rescale"],
)
def analyze(ctx, no_args_is_help=True, **kwargs):
    """\b
    Run the PyDamage analysis

    BAM: path to BAM/SAM/CRAM alignment file. MD tags need to be set.
    """
    pydamage_analyze(**kwargs, **ctx.obj)


@cli.command()
@click.pass_context
@click.argument("csv", type=click.Path(exists=True))
def filter(ctx, no_args_is_help=True, **kwargs):
    """\b
    Filter PyDamage results on predicted accuracy and qvalue thresholds.

    CSV: path to PyDamage result file
    """

    apply_filter(**kwargs, **ctx.obj)


@cli.command()
def cite():
    """Get pydamage citation in bibtex format"""
    get_citation()


if __name__ == "__main__":
    cli()



#!/usr/bin/env python3
from pydamage import utils
import multiprocessing
from functools import partial
from pydamage import damage
from pydamage.plot import damageplot
from pydamage.exceptions import PyDamageWarning, AlignmentFileError
from pydamage.accuracy_model import prepare_data, glm_predict
from pydamage.rescale import rescale_bam
from pydamage.models import glm_model_params
import os
from tqdm import tqdm
import warnings
from pydamage import __version__
from collections import ChainMap
import logging

def pydamage_analyze(
    bam,
    wlen=30,
    minlen=0,
    threshold=0.5,
    show_al=False,
    process=1,
    outdir="",
    plot=False,
    verbose=False,
    force=False,
    group=False,
    rescale=False,
    subsample=None,
    no_ga=False,
):
    """Runs the pydamage analysis for each reference separately

    Args:
        bam(str): Path to alignment (sam/bam/cram) file
        minlen(int): minimum reference sequence length threshold
        threshold(float): Predicted accuracy threshold
        wlen(int): window length
        show_al(bool): print alignments representations
        process(int):  Number of  processes for parellel computing
        outdir(str): Path to output directory
        verbose(bool): verbose mode
        force(bool): force overwriting of results directory
        group(bool): Use entire BAM file as single reference for analysis
        plot(bool): Write damage fitting plots to disk
        rescale(bool): Rescale base quality scores using the PyDamage damage model
        subsample(float): Subsample a fraction of the reads for damage modelling
        no_ga(bool): Do not use G->A transitions
    Returns:
        pd.DataFrame: pandas DataFrame containg pydamage results

    """
    g2a = not no_ga
    if subsample and rescale:
        raise ValueError("Cannot use subsample and rescale together")

    if no_ga and rescale:
        logging.warning("G to A transitions will not be used for rescaling !")

    if verbose:
        logging.info(f"Pydamage version {__version__}\n")
    utils.makedir(outdir, force=force)

    refs, mode = utils.prepare_bam(bam, minlen=minlen)

    proc = max(1, min(len(refs), process))

    ##########################
    # Simple loop for debugging
    ##########################
    # filt_res = []
    # for ref in refs:
    #     res = damage.test_damage(
    #         bam=bam,
    #         ref=ref,
    #         wlen=wlen,
    #         show_al=show_al,
    #         mode=mode,
    #         process=process,
    #         verbose=verbose,
    #     )
    #     if res:
    #         filt_res.append(res)
    #     break
    ######################
    # Multiprocessing code
    ######################

    test_damage_partial = partial(
        damage.test_damage,
        bam=bam,
        mode=mode,
        wlen=wlen,
        g2a=g2a,
        subsample=subsample,
        show_al=show_al,
        process=process,
        verbose=verbose,
    )
    if group:
        logging.info("Estimating and testing Damage")
        filt_res, read_dict = damage.test_damage(
            ref=None,
            bam=bam,
            mode=mode,
            wlen=wlen,
            g2a=g2a,
            subsample=subsample,
            show_al=show_al,
            process=process,
            verbose=verbose,
        )
        filt_res = [filt_res]
    else:
        if len(refs) > 0:
            with multiprocessing.Pool(proc) as p:
                res = list(
                    tqdm(
                        p.imap(test_damage_partial, refs),
                        total=len(refs),
                        desc="Estimating and testing Damage",
                    ),
                )
            filt_res, read_dicts = zip(*res)
            filt_res = [i for i in filt_res if i]
            read_dicts = [i for i in read_dicts if i]
            read_dict = dict(ChainMap(*read_dicts))
        else:
            raise AlignmentFileError(
                "No reference sequences were found in alignment file"
            )

    ######################
    ######################

    if not group:
        logging.info(f"{len(filt_res)} contig(s) analyzed by Pydamage")
    if len(filt_res) == 0:
        warnings.warn(
            "No alignments were found, check your alignment file", PyDamageWarning
        )

    if plot and len(filt_res) > 0:
        logging.info("Generating Pydamage plots")
        plotdir = f"{outdir}/plots"
        utils.makedir(plotdir, confirm=False)

        plot_partial = partial(damageplot, outdir=plotdir, wlen=wlen, plot_g2a=g2a)
        with multiprocessing.Pool(proc) as p:
            list(tqdm(p.imap(plot_partial, filt_res), total=len(filt_res)))
    df_pydamage = utils.pandas_processing(res_dict=filt_res, wlen=wlen, g2a=g2a)

    prep_df_glm = prepare_data(df_pydamage)
    df_glm = glm_predict(prep_df_glm, glm_model_params)

    df = df_glm.merge(df_pydamage, left_index=True, right_index=True)
    utils.df_to_csv(df, outdir)
    
    if rescale:
        rescale_bam(
            bam=bam,
            threshold=threshold,
            alpha=0.05,
            wlen=wlen,
            damage_dict=df.to_dict(),
            read_dict=read_dict,
            grouped=group,
            outname=os.path.join(outdir, "pydamage_rescaled.bam"),
            threads=process,
        )
    
    return df



import pysam
import numpy as np
from array import array
from pydamage.models import damage_model
from pydamage import __version__
from tqdm import tqdm
from numba import jit
import sys


@jit(
    cache=True,
    fastmath=True,
)
def phred_to_prob(qual):
    """Convert Phred quality score to probability

    Args:
        qual (array): Array of unsigned integer Phred quality scores

    Returns:
        np.array(int): Array of read error probabilities
    """
    return 10 ** (-qual / 10)


@jit(
    cache=True,
    fastmath=True,
)
def compute_new_prob(e, d):
    """Compute new probability of base calling  error accounting for ancient damage

    Args:
        e (np.array): Array of read error probabilities
        d (np.array): Array of damage probabilities
    Returns:
        np.array(int): Array of new read error probabilities
    """
    return np.round(-10 * np.log10(1 - np.multiply(1 - e, 1 - d)), 0).astype(np.int64)


def rescale_qual(read_qual, dmg_pmf, damage_bases, reverse):
    """Rescale quality scores using damage model
    Args:
        read_qual (array): Array of Phred quality scores
        dmg_pmf (array): Array of damage model probabilities
        damage_bases (array): Array of positions with damage
        reverse(bool): Read mapped to reverse strand
    Returns:
        np.array(int): Array of rescaled Phred quality scores
    """
    e = phred_to_prob(np.array(read_qual).astype(np.int64))
    if reverse:
        e = e[::-1]
    d = np.zeros(len(read_qual))
    d[damage_bases] = dmg_pmf[damage_bases]
    r = compute_new_prob(e, d)
    if reverse:
        r = r[::-1]
    return r


def rescale_bam(
    bam, threshold, alpha, wlen, damage_dict, read_dict, grouped, outname, threads
):
    """Rescale quality scores in BAM file using damage model

    Args:
        bam (str): Path to BAM file
        threshold (float): Predicted accuracy threshold
        alpha (float): Q-value threshold
        wlen (int): Window size for damage model
        damage_dict (dict): Damage model parameters
        read_dict (dict): Dictionary of read names
        grouped (bool): Grouped analysis
        outname (str): Path to output BAM file
        threads(int): Number of threads
    """
    with pysam.AlignmentFile(bam, "rb", threads=threads) as al:
        hd = al.header.to_dict()
        hd["PG"].append(
            {
                "ID": "pydamage",
                "PN": "pydamage",
                "VN": __version__,
                "CL": " ".join(sys.argv),
            }
        )
        refs = al.references
        with pysam.AlignmentFile(outname, "wb", threads=threads, header=hd) as out:
            for ref in tqdm(refs, desc="Rescaling quality scores"):
                if grouped:
                    pydam_ref = "reference"
                else:
                    pydam_ref = ref
                dmg = damage_model()
                if pydam_ref in read_dict:
                    if (
                        threshold
                        and 
                        threshold <= damage_dict["predicted_accuracy"][pydam_ref]
                    ) and (alpha and alpha >= damage_dict["qvalue"][pydam_ref]):
                        dmg_pmf = dmg.fit(
                            x=np.arange(wlen),
                            p=damage_dict["damage_model_p"][pydam_ref],
                            pmin=damage_dict["damage_model_pmin"][pydam_ref],
                            pmax=damage_dict["damage_model_pmax"][pydam_ref],
                        )
                        for read in al.fetch(ref):
                            if read.query_name in read_dict[pydam_ref]:
                                qual = read.query_qualities
                                read.query_sequence = read.query_sequence
                                read.query_qualities = array(
                                    "B",
                                    rescale_qual(
                                        qual,
                                        dmg_pmf,
                                        read_dict[pydam_ref][read.query_name],
                                        reverse=read.is_reverse,
                                    ),
                                )
                            out.write(read)
                    else:
                        for read in al.fetch(ref):
                            out.write(read)
                else:
                    for read in al.fetch(ref):
                        out.write(read)



import numpy as np
from tqdm import tqdm
import pysam
from numba import jit
from functools import partial
import multiprocessing
import pandas as pd
from io import StringIO
import logging


@jit()
def compute_coverage_sum(pysam_cov):
    """Computes average coverage of a reference

    Args:
        pysam_cov (np.array): Four dimensional array of coverage for each base
        ref_len (int): length of reference
    Returns:
        float: mean coverage of reference
    """
    return (
        np.sum(
            np.sum(pysam_cov, axis=0)
        )
    )

def _get_contig_stats(ref, al_handle):
    """
    Get the contig stats for a given reference and contig.
    """
    reflen = al_handle.get_reference_length(ref)
    cov = compute_coverage_sum(
        np.array(al_handle.count_coverage(contig=ref))
    )
    nb_reads_aligned = al_handle.count(contig=ref)
    

    return reflen, cov, nb_reads_aligned

def get_contig_stats(al_handle, ref):
    reflen, cov_sum, nb_reads_aligned = _get_contig_stats(ref, al_handle)
    
    cov = cov_sum/reflen if reflen > 0 else 0
    
    return reflen, cov, nb_reads_aligned

def get_ref_stats(bam):
    cov_stats = pysam.coverage(bam)
    df = pd.read_table(StringIO(cov_stats))
    df = df.query("numreads > 0")
    cov = (df['meandepth']*df['endpos']).sum()/df['endpos'].sum()
    reflen = df['endpos'].sum()
    nb_reads_aligned = df['numreads'].sum()

    return reflen, cov, nb_reads_aligned



def get_ref_stats_pysam(al_handle):
    cov_sum = np.empty((0,1), dtype="uint32")
    nb_reads_aligned = np.empty((0,1), dtype="uint32")
    reflen = np.empty((0,1), dtype="uint32")

    for r in tqdm(al_handle.references):
        _reflen, _cov, _nb_reads_aligned = _get_contig_stats(r, al_handle)
        if _nb_reads_aligned > 0:
            nb_reads_aligned = np.append(nb_reads_aligned, _nb_reads_aligned)
            reflen = np.append(reflen, _reflen)
            cov_sum = np.append(cov_sum, _cov)
    
    cov = np.sum(cov_sum)/np.sum(reflen) if np.sum(reflen) > 0 else 0
    
    return np.sum(reflen), cov, np.sum(nb_reads_aligned)

def get_ref_stats_multi(bam, process):
    
    get_contig_stats_partial = partial(_get_contig_stats, bam=bam, process=process)
    al_handle = pysam.AlignmentFile(bam, mode="rb", threads=process)
    refs = al_handle.references
    al_handle.close()
    with multiprocessing.Pool(process) as p:
        res = np.array(p.map(get_contig_stats_partial, refs))
    return np.sum(res[:, 0]), np.sum(res[:, 1])/np.sum(res[:, 0]), np.sum(res[:, 2])


#!/usr/bin/env python3

import pysam
from pydamage.parse_damage import damage_al
from pydamage.model_fit import fit_models
from pydamage.contig_stats import get_contig_stats, get_ref_stats
from pydamage import models
import numpy as np
from tqdm import tqdm
import logging

def sort_count_array_dict(int_array):
    """Sorts and counts unique values in an array

    Args:
        int_array (np.array): Array of integers
    Returns:
        np.array: sorted unique values
        np.array: counts of unique values
    """
    pos, counts = np.unique(int_array, return_counts=True)
    return pos, counts

class al_to_damage:
    def __init__(self, reference, al_handle, wlen, g2a, subsample):
        """Constructor of the class

        Args:
            reference (string): a reference from the indexed bam file
            al_handle(pysam.AlignmentFile)
            wlen (int): window length
            g2a (bool): use GtoA transitions
            subsample(float): subsample reads
        """
        self.alignments = al_handle.fetch(reference)
        self.reference = reference
        self.wlen = wlen
        self.g2a = g2a
        self.subsample = subsample

    def __repr__(self):
        return(f"Reference {self.reference}")

    def get_damage(self, show_al):
        """Compute CtoT substitutions

        Args:
            show_al(bool): print alignments representations
        Returns:
            list: C - positions from 5' of C in reads
            list: CT- positions from 5' of CtoT transitions in reads
            list: G - positions from 3' of G in reverse reads
            list: GA- positions from 3' of GtoA transitions in reverse reads
            list: damage_bases - positions of C2T and G2A bases (if reversed)
            list: C_G_bases - positions of C2C and G2G bases (if reversed)
        """
        self.C = []
        self.CT = []
        self.damage_bases = []
        self.C_G_bases = []
        self.G = []
        self.GA = []
        self.no_mut = []
        self.read_dict = {self.reference: dict()}
        if self.reference is None:
            iterator = tqdm(self.alignments, desc="Compute damage for entire reference")
        else:
            iterator = self.alignments
        for al in iterator:
            if al.is_unmapped is False and (
                not self.subsample or np.random.rand() < self.subsample
            ):
                all_damage = damage_al(
                    reference=al.get_reference_sequence(),
                    read_name=al.query_name,
                    is_reverse=al.is_reverse,
                    ref_name=self.reference,
                    query=al.query_sequence,
                    cigartuple=al.cigartuples,
                    wlen=self.wlen,
                    g2a=self.g2a,
                    show_al=show_al,
                )
                self.C += all_damage["C"]
                self.CT += all_damage["CT"]
                self.G += all_damage["G"]
                self.GA += all_damage["GA"]
                self.C_G_bases += all_damage["C"]
                CT_GA = all_damage["CT"]
                if self.g2a:
                    CT_GA += all_damage["GA"]
                    self.C_G_bases += all_damage["G"]
                self.damage_bases += CT_GA
                self.no_mut += all_damage["no_mut"]
                if len(CT_GA) > 0 and (
                    (al.is_reverse and self.g2a) or (not al.is_reverse)
                ):
                    self.read_dict[self.reference][al.query_name] = np.array(CT_GA)

    def compute_damage(self):
        """Computes the amount of damage for statistical modelling"""

        # All C in reference
        C_pos, C_counts = sort_count_array_dict(np.array(self.C, dtype="uint32"))

        # All G in reference
        G_pos, G_counts = sort_count_array_dict(np.array(self.G, dtype="uint32"))

        # CtoT transitions
        CT_pos, CT_counts = sort_count_array_dict(np.array(self.CT, dtype="uint32"))

        # GtoA transitions
        GA_pos, GA_counts = sort_count_array_dict(np.array(self.GA, dtype="uint32"))

        # All transitions
        damage_bases_pos, damage_bases_counts = sort_count_array_dict(np.array(self.damage_bases, dtype="uint32"))

        # All C and G in reference
        C_G_bases_pos, C_G_bases_counts = sort_count_array_dict(np.array(self.C_G_bases, dtype="uint32"))

        # All conserved C and G
        no_mut_pos, no_mut_counts = sort_count_array_dict(np.array(self.no_mut, dtype="uint32"))

        CT_damage_amount = np.zeros(self.wlen)
        CT_damage_amount[CT_pos] = CT_counts / C_counts[CT_pos]

        GA_damage_amount = np.zeros(self.wlen)
        GA_damage_amount[GA_pos] = GA_counts / G_counts[GA_pos]

        damage_amount = np.zeros(self.wlen)
        damage_amount[damage_bases_pos] = damage_bases_counts / C_G_bases_counts[damage_bases_pos]

        _ = np.zeros(self.wlen, dtype="uint32")
        _[damage_bases_pos] = damage_bases_counts
        damage_bases_counts = _

        _ = np.zeros(self.wlen, dtype="uint32")
        _[no_mut_pos] = no_mut_counts
        no_mut_counts = _

        return (
            damage_bases_counts,  # Number of CtoT and GtoA per position
            no_mut_counts,  # Number of conserved C and G per position
            CT_damage_amount,
            GA_damage_amount,
            damage_amount,
        )





def check_model_fit(model_dict, wlen, verbose):
    """Check if model fitting makes sense

    Args:
        model_dict (dict): Dictionary containing LR test results
        wlen (int): window length
        verbose (bool): Run in verbose mode
    Returns:
        bool or model_dict(dict): False if test fails, model_dict otherwise
    """
    # Check that no fitted parameters or stdev are infinite
    if np.isinf(np.array(model_dict["model_params"])).any():
        if verbose:
            logging.warning(f"Unreliable model fit for {model_dict['reference']}")
        return False

    return model_dict


def test_damage(ref, bam, mode, wlen, g2a, subsample, show_al, process, verbose):
    """Prepare data and run LRtest to test for damage

    Args:
        ref (str): name of reference in alignment file
        bam (str): bam file
        mode (str): opening mode of alignment file
        wlen (int): window length
        g2a (bool): Use GtoA transitions
        subsample (float): Subsample reads
        show_al (bool): Show alignment representations
        process (int): Number of process for parallelization
        verbose (bool): Run in verbose mode
    Returns:
        dict: Dictionary containing LR test results
    """
    al_handle = pysam.AlignmentFile(bam, mode=mode, threads=process)
    try:
        if ref is None:
            logging.info("Computing alignment stats for entire reference")
            reflen, cov, nb_reads_aligned = get_ref_stats(bam)
            refname = "reference"
        else:
            reflen, cov, nb_reads_aligned = get_contig_stats(al_handle, ref)
            refname = ref

        if subsample:
            cov = cov * subsample
            nb_reads_aligned = np.round(nb_reads_aligned * subsample, 0)

        al = al_to_damage(
            reference=ref, al_handle=al_handle, wlen=wlen, g2a=g2a, subsample=subsample
        )
        al.get_damage(show_al=show_al)
        read_dict = al.read_dict
        if len(read_dict.keys()) == 1 and list(read_dict.keys())[0] is None:
            read_dict[refname] = read_dict.pop(None)
        (
            mut_count,
            conserved_count,
            CT_damage,
            GA_damage,
            all_damage,
        ) = al.compute_damage()

        al_handle.close()

        model_A = models.damage_model()
        model_B = models.null_model()
        test_res = fit_models(
            ref=ref,
            model_A=model_A,
            model_B=model_B,
            damage=all_damage,
            mut_count=mut_count,
            conserved_count=conserved_count,
            verbose=verbose,
        )
        test_res["reference"] = refname
        test_res["nb_reads_aligned"] = nb_reads_aligned
        test_res["coverage"] = cov
        test_res["reflen"] = reflen

        CT_log = {}
        GA_log = {}

        for i in range(wlen):
            CT_log[f"CtoT-{i}"] = CT_damage[i]
            GA_log[f"GtoA-{i}"] = GA_damage[i]
        test_res.update(CT_log)
        test_res.update(GA_log)
        return (check_model_fit(test_res, wlen, verbose), read_dict)

    except (ValueError, RuntimeError) as e:
        if verbose:
            logging.warning(f"Model fitting for {ref} failed")
            logging.warning(f"Model fitting error: {e}")
            logging.warning(
                f"nb_reads_aligned: {nb_reads_aligned} - coverage: {cov}"
                f" - reflen: {reflen}\n"
            )
        al_handle.close()
        return False



def get_citation():
    BIB = """
@article{borry_pydamage_2021,
    author = {Borry, Maxime and Hübner, Alexander and Rohrlach, Adam B. and Warinner, Christina},
    doi = {10.7717/peerj.11845},
    issn = {2167-8359},
    journal = {PeerJ},
    language = {en},
    month = {July},
    note = {Publisher: PeerJ Inc.},
    pages = {e11845},
    shorttitle = {PyDamage},
    title = {PyDamage: automated ancient damage identification and estimation for contigs in ancient DNA de novo assembly},
    url = {https://peerj.com/articles/11845},
    urldate = {2021-07-27},
    volume = {9},
    year = {2021}
}
"""
    print(BIB)


from numpy import exp


def prepare_data(pd_df):
    """Prepare pydamage result data for accuracy modelling

    Args:
        pd_df (pandas DataFrame):pydamage df result
    """
    pd_df = pd_df[["coverage", "reflen", "damage_model_pmax"]].rename(
        columns={
            "damage_model_pmax": "damage",
            "reflen": "contiglength",
            "coverage": "actual_cov",
        }
    )
    pd_df = pd_df.astype({"actual_cov": float, "contiglength": int, "damage": float})
    return pd_df


def glm_predict(df, model_params):
    """Predict accuracy

    Args:
        df (pandas DataFrame): prepared pydamage results
        model_params (dict): model parameters
    """
    df["predicted_accuracy"] = (
        model_params["intercept"]
        + model_params["actual_cov"] * df["actual_cov"]
        + model_params["damage"] * df["damage"]
        + model_params["contiglength"] * df["contiglength"]
    )
    df["predicted_accuracy"] = 1 / (1 + exp(-df["predicted_accuracy"]))
    return df["predicted_accuracy"].to_frame()
